@@ -31,8 +31,6 @@ pub struct RosomaxaConfig {
     pub learning_rate: f64,
     /// A node rebalance memory of GSOM.
     pub rebalance_memory: usize,
-    /// A rebalance count.
-    pub rebalance_count: usize,
     /// A ratio of exploration phase.
     pub exploration_ratio: f64,
 }
@@ -45,12 +43,11 @@ impl RosomaxaConfig {
             selection_size,
             elite_size: 2,
             node_size: 2,
-            spread_factor: 0.25,
-            distribution_factor: 0.25,
+            spread_factor: 0.75,
+            distribution_factor: 0.75,
             objective_reshuffling: 0.01,
             learning_rate: 0.1,
             rebalance_memory: 100,
-            rebalance_count: 2,
             exploration_ratio: 0.9,
         }
     }
@@ -139,11 +136,10 @@ where
             RosomaxaPhases::Exploration { network, coordinates, selection_size, .. } => {
                 let (elite_explore_size, node_explore_size) = match *selection_size {
                     value if value > 6 => {
-                        let elite_size = self.environment.random.uniform_int(2, 4) as usize;
+                        let elite_size = self.environment.random.uniform_int(1, 2) as usize;
                         (elite_size, 2)
                     }
-                    value if value > 4 => (2, 2),
-                    value if value > 2 => (2, 1),
+                    value if value > 4 => (1, 2),
                     _ => (1, 1),
                 };
 
@@ -180,6 +176,19 @@ where
         self.elite.ranked()
     }
 
+    fn all<'a>(&'a self) -> Box<dyn Iterator<Item = &Self::Individual> + 'a> {
+        match &self.phase {
+            RosomaxaPhases::Exploration { network, .. } => {
+                Box::new(self.elite.all().chain(network.get_nodes().flat_map(|node| {
+                    // NOTE see above
+                    let node = node.read().unwrap();
+                    unsafe { &*(&node.storage.population as *const Elitism<O, S>) as &Elitism<O, S> }.all()
+                })))
+            }
+            _ => self.elite.all(),
+        }
+    }
+
     fn size(&self) -> usize {
         self.elite.size()
     }
@@ -217,8 +226,8 @@ where
 
     fn update_phase(&mut self, statistics: &HeuristicStatistics) {
         let selection_size = match statistics.speed {
-            HeuristicSpeed::Slow(ratio) => (self.config.selection_size as f64 * ratio).max(1.).round() as usize,
-            HeuristicSpeed::Moderate => self.config.selection_size,
+            HeuristicSpeed::Unknown | HeuristicSpeed::Moderate { .. } => self.config.selection_size,
+            HeuristicSpeed::Slow { ratio, .. } => (self.config.selection_size as f64 * ratio).max(1.).round() as usize,
         };
 
         match &mut self.phase {
@@ -247,8 +256,8 @@ where
                 selection_size: old_selection_size,
             } => {
                 let exploration_ratio = match old_statistics.speed {
-                    HeuristicSpeed::Slow(ratio) => self.config.exploration_ratio * ratio,
-                    HeuristicSpeed::Moderate => self.config.exploration_ratio,
+                    HeuristicSpeed::Unknown | HeuristicSpeed::Moderate { .. } => self.config.exploration_ratio,
+                    HeuristicSpeed::Slow { ratio, .. } => self.config.exploration_ratio * ratio,
                 };
 
                 if statistics.termination_estimate < exploration_ratio {
@@ -263,14 +272,13 @@ where
                         statistics,
                         best_fitness.as_slice(),
                         self.config.rebalance_memory,
-                        self.config.rebalance_count,
+                        self.config.learning_rate,
                     );
 
                     Self::fill_populations(
                         network,
                         coordinates,
                         best_fitness.as_slice(),
-                        statistics,
                         self.environment.random.as_ref(),
                     );
                 } else {
@@ -291,7 +299,6 @@ where
         network: &'a IndividualNetwork<O, S>,
         coordinates: &mut Vec<(Coordinate, f64, usize)>,
         best_fitness: &[f64],
-        statistics: &HeuristicStatistics,
         random: &(dyn Random + Send + Sync),
     ) {
         coordinates.clear();
@@ -299,7 +306,7 @@ where
             let node = node.read().unwrap();
             let coordinate = node.storage.population.select().next().map(|individual| {
                 (
-                    coordinate.clone(),
+                    *coordinate,
                     relative_distance(best_fitness.iter().cloned(), individual.get_fitness()),
                     node.get_last_hits(network.get_current_time()),
                 )
@@ -308,34 +315,7 @@ where
             coordinate
         }));
 
-        let shuffle_amount = Self::calculate_shuffle_amount(statistics, coordinates.len());
-        if shuffle_amount != coordinates.len() {
-            // partially randomize order
-            if random.is_head_not_tails() {
-                coordinates.sort_by(|(_, distance_a, _), (_, distance_b, _)| compare_floats(*distance_a, *distance_b));
-            } else {
-                coordinates.sort_by(|(_, _, last_hit_a), (_, _, last_hit_b)| last_hit_a.cmp(last_hit_b));
-            }
-
-            coordinates.partial_shuffle(&mut random.get_rng(), shuffle_amount);
-        } else {
-            coordinates.shuffle(&mut random.get_rng());
-        }
-    }
-
-    fn calculate_shuffle_amount(statistics: &HeuristicStatistics, length: usize) -> usize {
-        let ratio = match statistics.improvement_1000_ratio {
-            v if v > 0.5 => {
-                // https://www.wolframalpha.com/input/?i=plot+0.66+*+%281-+1%2F%281%2Be%5E%28-10+*%28x+-+0.5%29%29%29%29%2C+x%3D0+to+1
-                let progress = statistics.termination_estimate;
-                let ratio = 0.5 * (1. - 1. / (1. + std::f64::consts::E.powf(-10. * (progress - 0.5))));
-                ratio.clamp(0.1, 0.5)
-            }
-            v if v > 0.2 => 0.5,
-            _ => 1.,
-        };
-
-        (length as f64 * ratio).round() as usize
+        coordinates.shuffle(&mut random.get_rng());
     }
 
     fn optimize_network(
@@ -343,48 +323,52 @@ where
         statistics: &HeuristicStatistics,
         best_fitness: &[f64],
         rebalance_memory: usize,
-        rebalance_count: usize,
+        init_learning_rate: f64,
     ) {
-        let rebalance_memory = rebalance_memory as f64;
-        let keep_size = match statistics.improvement_1000_ratio {
-            v if v > 0.2 => {
-                // https://www.wolframalpha.com/input/?i=plot+%281+-+1%2F%281%2Be%5E%28-10+*%28x+-+0.5%29%29%29%29%2C+x%3D0+to+1
-                let x = statistics.termination_estimate.clamp(0., 1.);
-                let ratio = 1. - 1. / (1. + std::f64::consts::E.powf(-10. * (x - 0.5)));
-                rebalance_memory + rebalance_memory * ratio
-            }
-            v if v > 0.1 => 2. * rebalance_memory,
-            v if v > 0.01 => 3. * rebalance_memory,
-            _ => 4. * rebalance_memory,
-        } as usize;
+        // https://www.wolframalpha.com/input?i=plot+2+*+%281+-+1%2F%281%2Be%5E%28-10+*%28x+-+0.5%29%29%29%29%2C+x%3D0+to+1
+        let x = match statistics.improvement_1000_ratio {
+            v if v < 0.25 => v,
+            _ => statistics.termination_estimate,
+        }
+        .clamp(0., 1.);
 
-        if statistics.generation == 0 || network.size() <= keep_size {
+        let ratio = 2. * (1. - 1. / (1. + std::f64::consts::E.powf(-10. * (x - 0.5))));
+        let keep_size = rebalance_memory + (rebalance_memory as f64 * ratio) as usize;
+
+        if statistics.generation % rebalance_memory == 0 {
+            network.smooth(1);
+        }
+
+        if network.size() <= keep_size {
             return;
         }
 
+        if init_learning_rate < 1. {
+            network.set_learning_rate(statistics.termination_estimate.clamp(init_learning_rate, 1.));
+        }
+
         let get_distance = |node: &NodeLink<IndividualInput<S>, IndividualStorage<O, S>>| {
             let node = node.read().unwrap();
-            let individual = node.storage.population.select().next();
+            let individual = node.storage.population.ranked().next();
 
-            individual.map(|individual| relative_distance(best_fitness.iter().cloned(), individual.get_fitness()))
+            individual.map(|(individual, _)| relative_distance(best_fitness.iter().cloned(), individual.get_fitness()))
         };
 
         // determine percentile value
         let mut distances = network.get_nodes().filter_map(get_distance).collect::<Vec<_>>();
         distances.sort_by(|a, b| compare_floats(*b, *a));
-        let percentile_idx = if distances.len() > keep_size {
-            distances.len() - keep_size
-        } else {
-            // NOTE remove 75% of nodes
-            const PERCENTILE_THRESHOLD: f64 = 0.75;
 
-            (distances.len() as f64 * PERCENTILE_THRESHOLD) as usize
-        };
+        const PERCENTILE_THRESHOLD: f64 = 0.1;
+        let percentile_idx = (distances.len() as f64 * PERCENTILE_THRESHOLD) as usize;
 
         if let Some(distance_threshold) = distances.get(percentile_idx).cloned() {
-            network.retrain(rebalance_count, &|node| {
-                get_distance(node).map_or(false, |distance| distance < distance_threshold)
+            network.compact(&|node, unified_distance| {
+                // NOTE
+                // unified distance filter improves diversity property
+                // distance filter improves exploitation characteristic by removing old (or empty) nodes
+                unified_distance > 0.1 && get_distance(node).map_or(false, |distance| distance < distance_threshold)
             });
+            network.smooth(1);
         }
     }
 
@@ -418,6 +402,7 @@ where
                 rebalance_memory: config.rebalance_memory,
                 has_initial_error: true,
             },
+            environment.random.clone(),
             storage_factory,
         )
     }
@@ -439,6 +424,21 @@ where
     }
 }
 
+impl<'a, O, S> TryFrom<&'a Rosomaxa<O, S>> for NetworkState
+where
+    O: HeuristicObjective<Solution = S> + Shuffled,
+    S: HeuristicSolution + RosomaxaWeighted + DominanceOrdered,
+{
+    type Error = String;
+
+    fn try_from(value: &'a Rosomaxa<O, S>) -> Result<Self, Self::Error> {
+        match &value.phase {
+            RosomaxaPhases::Exploration { network, .. } => Ok(get_network_state(network)),
+            _ => Err("not in exploration state".to_string()),
+        }
+    }
+}
+
 #[allow(clippy::large_enum_variant)]
 enum RosomaxaPhases<O, S>
 where
@@ -502,7 +502,22 @@ where
     S: HeuristicSolution + RosomaxaWeighted + DominanceOrdered,
 {
     fn eval(&self) -> IndividualStorage<O, S> {
-        let mut elitism = Elitism::new(self.objective.clone(), self.random.clone(), self.node_size, self.node_size);
+        let mut elitism = Elitism::new_with_dedup(
+            self.objective.clone(),
+            self.random.clone(),
+            self.node_size,
+            self.node_size,
+            Box::new(|a, b| {
+                // TODO return slice instead of vector?
+                let weights_a = a.weights();
+                let weights_b = b.weights();
+
+                let distance = relative_distance(weights_a.iter().cloned(), weights_b.iter().cloned());
+
+                // NOTE custom dedup rule to increase diversity property
+                distance < 0.1
+            }),
+        );
         if self.random.is_hit(self.reshuffling_probability) {
             elitism.shuffle_objective();
         }
