@@ -1,5 +1,6 @@
 use super::*;
 use crate::example::*;
+use crate::helpers::example::create_example_objective;
 
 fn get_best_fitness(population: &Greedy<VectorObjective, VectorSolution>) -> f64 {
     population.objective.fitness(population.ranked().next().unwrap().0)
@@ -7,7 +8,7 @@ fn get_best_fitness(population: &Greedy<VectorObjective, VectorSolution>) -> f64
 
 #[test]
 fn can_keep_best_solution() {
-    let objective = Arc::new(VectorObjective::new(create_rosenbrock_function()));
+    let objective = create_example_objective();
     let mut population = Greedy::<_, _>::new(objective.clone(), 1, None);
 
     population.add(VectorSolution::new(vec![-1., -1.], objective.clone()));
